@@ -85,12 +85,12 @@ struct SectionFragment {
 // memory.
 struct SymbolAux {
   i32 got_idx = -1;
-  i32 gotplt_idx = -1;
   i32 gottp_idx = -1;
   i32 tlsgd_idx = -1;
   i32 tlsdesc_idx = -1;
   i32 plt_idx = -1;
   i32 pltgot_idx = -1;
+  i32 opd_idx = -1;
   i32 dynsym_idx = -1;
 };
 
@@ -114,6 +114,7 @@ public:
   }
 
   static constexpr i64 alignment = 4;
+
   OutputSection<E> &output_section;
   i32 thunk_idx = -1;
   i64 offset = -1;
@@ -239,6 +240,40 @@ struct InputSectionExtras<E> {
   std::vector<i32> r_deltas;
 };
 
+typedef enum { NONE, ERROR, COPYREL, PLT, CPLT, DYNREL, BASEREL } ScanAction;
+
+// This is a decision table for absolute relocations that is smaller
+// than the word size (e.g. R_X86_64_32). Since the dynamic linker
+// generally does not support dynamic relocations smaller than the
+// word size, we need to report an error if a relocation cannot be
+// resolved at link-time.
+constexpr static ScanAction absrel_table[3][4] = {
+  // Absolute  Local    Imported data  Imported code
+  {  NONE,     ERROR,   ERROR,         ERROR },  // Shared object
+  {  NONE,     ERROR,   ERROR,         ERROR },  // Position-independent exec
+  {  NONE,     NONE,    COPYREL,       CPLT  },  // Position-dependent exec
+};
+
+// This is a decision table for absolute relocations for the word
+// size data (e.g. R_X86_64_64). Unlike the absrel_table, we can emit
+// a dynamic relocation if we cannot resolve an address at link-time.
+constexpr static ScanAction dyn_absrel_table[3][4] = {
+  // Absolute  Local    Imported data  Imported code
+  {  NONE,     BASEREL, DYNREL,        DYNREL },  // Shared object
+  {  NONE,     BASEREL, DYNREL,        DYNREL },  // Position-independent exec
+  {  NONE,     NONE,    COPYREL,       CPLT   },  // Position-dependent exec
+};
+
+// This is for PC-relative relocations (e.g. R_X86_64_PC32).
+// We cannot promote them to dynamic relocations because the dynamic
+// linker generally does not support PC-relative relocations.
+constexpr static ScanAction pcrel_table[3][4] = {
+  // Absolute  Local    Imported data  Imported code
+  {  ERROR,    NONE,    ERROR,         PLT    },  // Shared object
+  {  ERROR,    NONE,    COPYREL,       PLT    },  // Position-independent exec
+  {  NONE,     NONE,    COPYREL,       CPLT   },  // Position-dependent exec
+};
+
 // InputSection represents a section in an input object file.
 template <typename E>
 class InputSection {
@@ -263,11 +298,13 @@ public:
   std::span<FdeRecord<E>> get_fdes() const;
   std::string_view get_func_name(Context<E> &ctx, i64 offset);
   bool is_relr_reloc(Context<E> &ctx, const ElfRel<E> &rel);
+  bool is_killed_by_icf() const;
 
   void record_undef_error(Context<E> &ctx, const ElfRel<E> &rel);
 
   ObjectFile<E> &file;
   OutputSection<E> *output_section = nullptr;
+  u64 sh_size = -1;
 
   std::string_view contents;
 
@@ -280,7 +317,6 @@ public:
   u32 shndx = -1;
   u32 relsec_idx = -1;
   u32 reldyn_offset = 0;
-  u32 sh_size = -1;
 
   // For COMDAT de-duplication and garbage collection
   std::atomic_bool is_alive = true;
@@ -288,24 +324,29 @@ public:
 
   bool address_significant : 1 = false;
   bool uncompressed : 1 = false;
-  bool killed_by_icf : 1 = false;
 
   // For garbage collection
   std::atomic_bool is_visited = false;
 
   // For ICF
+  //
+  // `leader` is the section that this section has been merged with.
+  // Three kind of values are possible:
+  // - `leader == nullptr`: This section was not eligible for ICF.
+  // - `leader == this`: This section was retained.
+  // - `leader != this`: This section was merged with another identical section.
   InputSection<E> *leader = nullptr;
   u32 icf_idx = -1;
   bool icf_eligible = false;
   bool icf_leaf = false;
 
 private:
-  void scan_abs_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel);
-  void scan_abs_dyn_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel);
-  void scan_pcrel_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel);
+  void scan_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel,
+                const ScanAction table[3][4]);
 
-  void apply_abs_dyn_rel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel,
-                         u8 *loc, u64 S, i64 A, u64 P, ElfRel<E> *&dynrel);
+  void apply_dyn_absrel(Context<E> &ctx, Symbol<E> &sym, const ElfRel<E> &rel,
+                        u8 *loc, u64 S, i64 A, u64 P, ElfRel<E> *&dynrel,
+                        const ScanAction table[3][4] = dyn_absrel_table);
 
   void copy_contents_riscv(Context<E> &ctx, u8 *buf);
 
@@ -325,9 +366,21 @@ void report_undef_errors(Context<E> &ctx);
 template <typename E>
 u64 get_eflags(Context<E> &ctx);
 
+template <typename E>
+i64 to_phdr_flags(Context<E> &ctx, Chunk<E> *chunk);
+
 template <typename E>
 bool is_relro(Context<E> &ctx, Chunk<E> *chunk);
 
+template <typename E>
+void write_plt_header(Context<E> &ctx, u8 *buf);
+
+template <typename E>
+void write_plt_entry(Context<E> &ctx, u8 *buf, Symbol<E> &sym);
+
+template <typename E>
+void write_pltgot_entry(Context<E> &ctx, u8 *buf, Symbol<E> &sym);
+
 typedef enum { HEADER, OUTPUT_SECTION, SYNTHETIC } ChunkKind;
 
 // Chunk represents a contiguous region in an output file.
@@ -346,7 +399,18 @@ public:
   std::string_view name;
   ElfShdr<E> shdr = {};
   i64 shndx = 0;
-  i64 extra_addralign = 1;
+
+  // Some synethetic sections add local symbols to the output.
+  // For example, range extension thunks adds function_name@thunk
+  // symbol for each thunk entry. The following members are used
+  // for such synthesizing symbols.
+  virtual void compute_symtab_size(Context<E> &ctx) {};
+  virtual void populate_symtab(Context<E> &ctx) {};
+
+  i64 local_symtab_idx = 0;
+  i64 num_local_symtab = 0;
+  i64 strtab_size = 0;
+  i64 strtab_offset = 0;
 
 protected:
   Chunk() { shdr.sh_addralign = 1; }
@@ -420,6 +484,9 @@ public:
   void copy_buf(Context<E> &ctx) override;
   void write_to(Context<E> &ctx, u8 *buf) override;
 
+  void compute_symtab_size(Context<E> &ctx) override;
+  void populate_symtab(Context<E> &ctx) override;
+
   std::vector<InputSection<E> *> members;
   u32 idx;
 
@@ -457,6 +524,11 @@ public:
     this->shdr.sh_type = SHT_PROGBITS;
     this->shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
     this->shdr.sh_addralign = sizeof(Word<E>);
+
+    // We always create a .got so that _GLOBAL_OFFSET_TABLE_ has
+    // something to point to. s390x psABI defines GOT[1] as a
+    // reserved slot, so we allocate one more on s390x.
+    this->shdr.sh_size = (is_s390x<E> ? 2 : 1) * sizeof(Word<E>);
   }
 
   void add_got_symbol(Context<E> &ctx, Symbol<E> *sym);
@@ -466,10 +538,13 @@ public:
   void add_tlsld(Context<E> &ctx);
 
   u64 get_tlsld_addr(Context<E> &ctx) const;
+  bool has_tlsld(Context<E> &ctx) const { return tlsld_idx != -1; }
   i64 get_reldyn_size(Context<E> &ctx) const;
-  void update_shdr(Context<E> &ctx) override;
   void copy_buf(Context<E> &ctx) override;
 
+  void compute_symtab_size(Context<E> &ctx) override;
+  void populate_symtab(Context<E> &ctx) override;
+
   std::vector<Symbol<E> *> got_syms;
   std::vector<Symbol<E> *> gottp_syms;
   std::vector<Symbol<E> *> tlsgd_syms;
@@ -491,10 +566,17 @@ public:
     this->shdr.sh_type = is_ppc<E> ? SHT_NOBITS : SHT_PROGBITS;
     this->shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
     this->shdr.sh_addralign = sizeof(Word<E>);
-    this->shdr.sh_size = sizeof(Word<E>) * (is_ppc<E> ? 2 : 3);
+    this->shdr.sh_size = HDR_SIZE;
   }
 
+  void update_shdr(Context<E> &ctx) override;
   void copy_buf(Context<E> &ctx) override;
+
+  static constexpr i64 HDR_SIZE =
+    (std::is_same_v<E, PPC64V2> ? 2 : 3) * sizeof(Word<E>);
+
+  static constexpr i64 ENTRY_SIZE =
+    (std::is_same_v<E, PPC64V1> ? 3 : 1) * sizeof(Word<E>);
 };
 
 template <typename E>
@@ -514,8 +596,12 @@ public:
   }
 
   void add_symbol(Context<E> &ctx, Symbol<E> *sym);
+  void update_shdr(Context<E> &ctx) override;
   void copy_buf(Context<E> &ctx) override;
 
+  void compute_symtab_size(Context<E> &ctx) override;
+  void populate_symtab(Context<E> &ctx) override;
+
   std::vector<Symbol<E> *> symbols;
 };
 
@@ -532,6 +618,9 @@ public:
   void add_symbol(Context<E> &ctx, Symbol<E> *sym);
   void copy_buf(Context<E> &ctx) override;
 
+  void compute_symtab_size(Context<E> &ctx) override;
+  void populate_symtab(Context<E> &ctx) override;
+
   std::vector<Symbol<E> *> symbols;
 };
 
@@ -779,7 +868,6 @@ public:
     this->name = is_relro ? ".copyrel.rel.ro" : ".copyrel";
     this->shdr.sh_type = SHT_NOBITS;
     this->shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
-    this->shdr.sh_addralign = 64;
   }
 
   void add_symbol(Context<E> &ctx, Symbol<E> *sym);
@@ -967,6 +1055,21 @@ private:
   std::vector<i64> offsets;
 };
 
+// PT_GNU_RELRO works on page granularity. We want to align its end to
+// a page boundary. We append this section at end of a segment so that
+// the segment always ends at a page boundary.
+template <typename E>
+class RelroPaddingSection : public Chunk<E> {
+public:
+  RelroPaddingSection() {
+    this->name = ".relro_padding";
+    this->shdr.sh_type = SHT_NOBITS;
+    this->shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
+    this->shdr.sh_addralign = 1;
+    this->shdr.sh_size = 1;
+  }
+};
+
 //
 // dwarf.cc
 //
@@ -1104,7 +1207,7 @@ public:
   void claim_unresolved_symbols(Context<E> &ctx);
   void scan_relocations(Context<E> &ctx);
   void convert_common_symbols(Context<E> &ctx);
-  void compute_symtab(Context<E> &ctx);
+  void compute_symtab_size(Context<E> &ctx);
   void populate_symtab(Context<E> &ctx);
 
   i64 get_shndx(const ElfSym<E> &esym);
@@ -1180,12 +1283,13 @@ public:
   void parse(Context<E> &ctx);
   void resolve_symbols(Context<E> &ctx) override;
   std::vector<Symbol<E> *> find_aliases(Symbol<E> *sym);
+  i64 get_alignment(Symbol<E> *sym);
   bool is_readonly(Context<E> &ctx, Symbol<E> *sym);
 
   void mark_live_objects(Context<E> &ctx,
                          std::function<void(InputFile<E> *)> feeder) override;
 
-  void compute_symtab(Context<E> &ctx);
+  void compute_symtab_size(Context<E> &ctx);
   void populate_symtab(Context<E> &ctx);
 
   bool is_needed = false;
@@ -1307,10 +1411,11 @@ collect_output_sections(Context<E> &);
 template <typename E> void compute_section_sizes(Context<E> &);
 template <typename E> void sort_output_sections(Context<E> &);
 template <typename E> void claim_unresolved_symbols(Context<E> &);
-template <typename E> void scan_rels(Context<E> &);
+template <typename E> void scan_relocations(Context<E> &);
 template <typename E> void construct_relr(Context<E> &);
 template <typename E> void create_output_symtab(Context<E> &);
 template <typename E> void create_reloc_sections(Context<E> &);
+template <typename E> void copy_chunks(Context<E> &);
 template <typename E> void apply_version_script(Context<E> &);
 template <typename E> void parse_symbol_version(Context<E> &);
 template <typename E> void compute_import_export(Context<E> &);
@@ -1334,6 +1439,47 @@ void sort_arm_exidx(Context<ARM32> &ctx);
 template <typename E>
 i64 riscv_resize_sections(Context<E> &ctx);
 
+//
+// arch-ppc64v1.cc
+//
+
+void ppc64v1_rewrite_opd(Context<PPC64V1> &ctx);
+void ppc64v1_scan_symbols(Context<PPC64V1> &ctx);
+
+class PPC64OpdSection : public Chunk<PPC64V1> {
+public:
+  PPC64OpdSection() {
+    this->name = ".opd";
+    this->shdr.sh_type = SHT_PROGBITS;
+    this->shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
+    this->shdr.sh_addralign = 8;
+  }
+
+  void add_symbol(Context<PPC64V1> &ctx, Symbol<PPC64V1> *sym);
+  void copy_buf(Context<PPC64V1> &ctx) override;
+
+  static constexpr i64 ENTRY_SIZE = sizeof(Word<PPC64V1>) * 3;
+
+  std::vector<Symbol<PPC64V1> *> symbols;
+};
+
+//
+// arch-s390x.cc
+//
+
+class S390XTlsGetOffsetSection : public Chunk<S390X> {
+public:
+  S390XTlsGetOffsetSection() {
+    this->name = ".tls_get_offset";
+    this->shdr.sh_type = SHT_PROGBITS;
+    this->shdr.sh_flags = SHF_ALLOC | SHF_EXECINSTR;
+    this->shdr.sh_addralign = 4;
+    this->shdr.sh_size = 28;
+  }
+
+  void copy_buf(Context<S390X> &ctx) override;
+};
+
 //
 // arch-sparc.cc
 //
@@ -1416,7 +1562,7 @@ struct Context {
     BuildId build_id;
     CetReportKind z_cet_report = CET_REPORT_NONE;
     CompressKind compress_debug_sections = COMPRESS_NONE;
-    SeparateCodeKind z_separate_code = SEPARATE_LOADABLE_SEGMENTS;
+    SeparateCodeKind z_separate_code = NOSEPARATE_CODE;
     ShuffleSectionsKind shuffle_sections = SHUFFLE_SECTIONS_NONE;
     UnresolvedKind unresolved_symbols = UNRESOLVED_ERROR;
     bool Bsymbolic = false;
@@ -1523,7 +1669,7 @@ struct Context {
 
   std::vector<VersionPattern> version_patterns;
   u16 default_version = VER_NDX_GLOBAL;
-  bool version_specified = false;
+  bool default_version_from_version_script = false; // true if default_version is set by a wildcard in version script.
   i64 page_size = -1;
 
   // Reader context
@@ -1609,7 +1755,14 @@ struct Context {
   NotePackageSection<E> *note_package = nullptr;
   NotePropertySection<E> *note_property = nullptr;
   GdbIndexSection<E> *gdb_index = nullptr;
+  RelroPaddingSection<E> *relro_padding = nullptr;
+  PPC64OpdSection *ppc64_opd = nullptr;
   SparcTlsGetAddrSection *sparc_tls_get_addr = nullptr;
+  S390XTlsGetOffsetSection *s390x_tls_get_offset = nullptr;
+
+  // Frequently accessed symbols
+  Symbol<E> *tls_get_addr = nullptr;
+  Symbol<E> *tls_get_offset = nullptr;
 
   // For --gdb-index
   Chunk<E> *debug_info = nullptr;
@@ -1628,7 +1781,6 @@ struct Context {
 
   u64 tls_begin = 0;
   u64 tp_addr = 0;
-  bool relax_tlsdesc = false;
 
   // Linker-synthesized symbols
   Symbol<E> *TOC = nullptr;
@@ -1689,6 +1841,7 @@ enum {
   NEEDS_TLSGD   = 1 << 4,
   NEEDS_COPYREL = 1 << 5,
   NEEDS_TLSDESC = 1 << 6,
+  NEEDS_OPD     = 1 << 7,
 };
 
 // A struct to hold taret-dependent symbol members.
@@ -1702,6 +1855,12 @@ struct SymbolExtras<E> {
   i32 thunk_sym_idx = -1;
 };
 
+// Flags for Symbol<E>::get_addr()
+enum {
+  NO_PLT = 1 << 0, // Request an address other than .plt
+  NO_OPD = 1 << 1, // Request an address other than .opd (PPC64V1 only)
+};
+
 // Symbol class represents a defined symbol.
 //
 // A symbol has not only one but several different addresses if it
@@ -1714,38 +1873,44 @@ public:
   Symbol(std::string_view name) : nameptr(name.data()), namelen(name.size()) {}
   Symbol(const Symbol<E> &other) : Symbol(other.name()) {}
 
-  u64 get_addr(Context<E> &ctx, bool allow_plt = true) const;
+  u64 get_addr(Context<E> &ctx, i64 flags = 0) const;
   u64 get_got_addr(Context<E> &ctx) const;
   u64 get_gotplt_addr(Context<E> &ctx) const;
   u64 get_gottp_addr(Context<E> &ctx) const;
   u64 get_tlsgd_addr(Context<E> &ctx) const;
   u64 get_tlsdesc_addr(Context<E> &ctx) const;
   u64 get_plt_addr(Context<E> &ctx) const;
+  u64 get_opd_addr(Context<E> &ctx) const;
 
   void set_got_idx(Context<E> &ctx, i32 idx);
-  void set_gotplt_idx(Context<E> &ctx, i32 idx);
   void set_gottp_idx(Context<E> &ctx, i32 idx);
   void set_tlsgd_idx(Context<E> &ctx, i32 idx);
   void set_tlsdesc_idx(Context<E> &ctx, i32 idx);
   void set_plt_idx(Context<E> &ctx, i32 idx);
   void set_pltgot_idx(Context<E> &ctx, i32 idx);
+  void set_opd_idx(Context<E> &ctx, i32 idx);
   void set_dynsym_idx(Context<E> &ctx, i32 idx);
 
   i32 get_got_idx(Context<E> &ctx) const;
-  i32 get_gotplt_idx(Context<E> &ctx) const;
   i32 get_gottp_idx(Context<E> &ctx) const;
   i32 get_tlsgd_idx(Context<E> &ctx) const;
   i32 get_tlsdesc_idx(Context<E> &ctx) const;
   i32 get_plt_idx(Context<E> &ctx) const;
   i32 get_pltgot_idx(Context<E> &ctx) const;
+  i32 get_opd_idx(Context<E> &ctx) const;
   i32 get_dynsym_idx(Context<E> &ctx) const;
 
   bool has_plt(Context<E> &ctx) const;
-  bool has_got(Context<E> &ctx) const;
+  bool has_got(Context<E> &ctx) const { return get_got_idx(ctx) != -1; }
+  bool has_gottp(Context<E> &ctx) const { return get_gottp_idx(ctx) != -1; }
+  bool has_tlsgd(Context<E> &ctx) const { return get_tlsgd_idx(ctx) != -1; }
+  bool has_tlsdesc(Context<E> &ctx) const { return get_tlsdesc_idx(ctx) != -1; }
+  bool has_opd(Context<E> &ctx) const { return get_opd_idx(ctx) != -1; }
 
   bool is_absolute() const;
   bool is_relative() const { return !is_absolute(); }
   bool is_local() const { return !is_imported && !is_exported; }
+  bool is_ifunc() const { return get_type() == STT_GNU_IFUNC; }
   bool is_remaining_undef_weak() const;
 
   InputSection<E> *get_input_section() const;
@@ -1834,7 +1999,7 @@ public:
   bool is_exported : 1 = false;
 
   // `is_canonical` is true if this symbol represents a "canonical" PLT.
-  // Here is the explanation as to what is the canonical PLT is.
+  // Here is the explanation as to what the canonical PLT is.
   //
   // In C/C++, the process-wide function pointer equality is guaratneed.
   // That is, if you take an address of a function `foo`, it's always
@@ -1898,8 +2063,8 @@ public:
   // The loader supports a feature so-called "copy relocations".
   // A copy relocation instructs the loader to copy data from a DSO to a
   // specified location in the main executable. By using this feature,
-  // you can make `foo`'s data to a BSS region at runtime. With that,
-  // you can apply relocations agianst `foo` as if `foo` existed in the
+  // we can copy `foo`'s data to a BSS region at runtime. With that,
+  // we can apply relocations agianst `foo` as if `foo` existed in the
   // main executable's BSS area, whose address is known at link-time.
   //
   // Copy relocations are used only by position-dependent executables.
@@ -2174,7 +2339,7 @@ InputSection<E>::get_tombstone(Symbol<E> &sym, SectionFragment<E> *frag) {
   // If the section was dead due to ICF, we don't want to emit debug
   // info for that section but want to set real values to .debug_line so
   // that users can set a breakpoint inside a merged section.
-  if (isec->killed_by_icf && s == ".debug_line")
+  if (isec->is_killed_by_icf() && s == ".debug_line")
     return {};
 
   // 0 is an invalid value in most debug info sections, so we use it
@@ -2191,6 +2356,11 @@ inline bool InputSection<E>::is_relr_reloc(Context<E> &ctx, const ElfRel<E> &rel
          (rel.r_offset % sizeof(Word<E>)) == 0;
 }
 
+template<typename E>
+inline bool InputSection<E>::is_killed_by_icf() const {
+  return this->leader && this->leader != this;
+}
+
 template <typename E>
 std::pair<SectionFragment<E> *, i64>
 MergeableSection<E>::get_fragment(i64 offset) {
@@ -2261,7 +2431,7 @@ inline InputSection<E> *ObjectFile<E>::get_section(const ElfSym<E> &esym) {
 }
 
 template <typename E>
-inline u64 Symbol<E>::get_addr(Context<E> &ctx, bool allow_plt) const {
+inline u64 Symbol<E>::get_addr(Context<E> &ctx, i64 flags) const {
   if (SectionFragment<E> *frag = get_frag()) {
     if (!frag->is_alive) {
       // This condition is met if a non-alloc section refers an
@@ -2280,8 +2450,12 @@ inline u64 Symbol<E>::get_addr(Context<E> &ctx, bool allow_plt) const {
       : ctx.copyrel->shdr.sh_addr + value;
   }
 
-  if (allow_plt && has_plt(ctx)) {
-    assert(is_imported || esym().st_type == STT_GNU_IFUNC);
+  if constexpr (std::is_same_v<E, PPC64V1>)
+    if (!(flags & NO_OPD) && has_opd(ctx))
+      return get_opd_addr(ctx);
+
+  if (!(flags & NO_PLT) && has_plt(ctx)) {
+    assert(is_imported || is_ifunc());
     return get_plt_addr(ctx);
   }
 
@@ -2290,7 +2464,7 @@ inline u64 Symbol<E>::get_addr(Context<E> &ctx, bool allow_plt) const {
     return value; // absolute symbol
 
   if (!isec->is_alive) {
-    if (isec->killed_by_icf)
+    if (isec->is_killed_by_icf())
       return isec->leader->get_addr() + value;
 
     if (isec->name() == ".eh_frame") {
@@ -2333,8 +2507,9 @@ inline u64 Symbol<E>::get_got_addr(Context<E> &ctx) const {
 
 template <typename E>
 inline u64 Symbol<E>::get_gotplt_addr(Context<E> &ctx) const {
-  assert(get_gotplt_idx(ctx) != -1);
-  return ctx.gotplt->shdr.sh_addr + get_gotplt_idx(ctx) * sizeof(Word<E>);
+  assert(get_plt_idx(ctx) != -1);
+  return ctx.gotplt->shdr.sh_addr + GotPltSection<E>::HDR_SIZE +
+         get_plt_idx(ctx) * GotPltSection<E>::ENTRY_SIZE;
 }
 
 template <typename E>
@@ -2362,6 +2537,13 @@ inline u64 Symbol<E>::get_plt_addr(Context<E> &ctx) const {
   return ctx.pltgot->shdr.sh_addr + get_pltgot_idx(ctx) * E::pltgot_size;
 }
 
+template <typename E>
+inline u64 Symbol<E>::get_opd_addr(Context<E> &ctx) const {
+  assert(get_opd_idx(ctx) != -1);
+  return ctx.ppc64_opd->shdr.sh_addr +
+         get_opd_idx(ctx) * PPC64OpdSection::ENTRY_SIZE;
+}
+
 template <typename E>
 inline void Symbol<E>::set_got_idx(Context<E> &ctx, i32 idx) {
   assert(aux_idx != -1);
@@ -2369,13 +2551,6 @@ inline void Symbol<E>::set_got_idx(Context<E> &ctx, i32 idx) {
   ctx.symbol_aux[aux_idx].got_idx = idx;
 }
 
-template <typename E>
-inline void Symbol<E>::set_gotplt_idx(Context<E> &ctx, i32 idx) {
-  assert(aux_idx != -1);
-  assert(ctx.symbol_aux[aux_idx].gotplt_idx < 0);
-  ctx.symbol_aux[aux_idx].gotplt_idx = idx;
-}
-
 template <typename E>
 inline void Symbol<E>::set_gottp_idx(Context<E> &ctx, i32 idx) {
   assert(aux_idx != -1);
@@ -2411,6 +2586,13 @@ inline void Symbol<E>::set_pltgot_idx(Context<E> &ctx, i32 idx) {
   ctx.symbol_aux[aux_idx].pltgot_idx = idx;
 }
 
+template <typename E>
+inline void Symbol<E>::set_opd_idx(Context<E> &ctx, i32 idx) {
+  assert(aux_idx != -1);
+  assert(ctx.symbol_aux[aux_idx].opd_idx < 0);
+  ctx.symbol_aux[aux_idx].opd_idx = idx;
+}
+
 template <typename E>
 inline void Symbol<E>::set_dynsym_idx(Context<E> &ctx, i32 idx) {
   assert(aux_idx != -1);
@@ -2423,11 +2605,6 @@ inline i32 Symbol<E>::get_got_idx(Context<E> &ctx) const {
   return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].got_idx;
 }
 
-template <typename E>
-inline i32 Symbol<E>::get_gotplt_idx(Context<E> &ctx) const {
-  return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].gotplt_idx;
-}
-
 template <typename E>
 inline i32 Symbol<E>::get_gottp_idx(Context<E> &ctx) const {
   return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].gottp_idx;
@@ -2453,6 +2630,11 @@ inline i32 Symbol<E>::get_pltgot_idx(Context<E> &ctx) const {
   return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].pltgot_idx;
 }
 
+template <typename E>
+inline i32 Symbol<E>::get_opd_idx(Context<E> &ctx) const {
+  return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].opd_idx;
+}
+
 template <typename E>
 inline i32 Symbol<E>::get_dynsym_idx(Context<E> &ctx) const {
   return (aux_idx == -1) ? -1 : ctx.symbol_aux[aux_idx].dynsym_idx;
@@ -2463,11 +2645,6 @@ inline bool Symbol<E>::has_plt(Context<E> &ctx) const {
   return get_plt_idx(ctx) != -1 || get_pltgot_idx(ctx) != -1;
 }
 
-template <typename E>
-inline bool Symbol<E>::has_got(Context<E> &ctx) const {
-  return get_got_idx(ctx) != -1;
-}
-
 template <typename E>
 inline bool Symbol<E>::is_absolute() const {
   if (file && file->is_dso)
@@ -2589,4 +2766,25 @@ inline bool is_c_identifier(std::string_view s) {
   return true;
 }
 
+template <typename E>
+inline bool relax_tlsgd(Context<E> &ctx, Symbol<E> &sym) {
+  return ctx.arg.relax && !ctx.arg.shared && !sym.is_imported;
+}
+
+template <typename E>
+inline bool relax_tlsld(Context<E> &ctx) {
+  return ctx.arg.relax && !ctx.arg.shared;
+}
+
+template <typename E>
+inline bool relax_tlsdesc(Context<E> &ctx, Symbol<E> &sym) {
+  // TLSDESC relocs must be always relaxed for statically-linked
+  // executables even if -no-relax is given. It is because a
+  // statically-linked executable doesn't contain a tranpoline
+  // function needed for TLSDESC.
+  if (ctx.arg.is_static)
+    return true;
+  return ctx.arg.relax && !ctx.arg.shared && !sym.is_imported;
+}
+
 } // namespace mold::elf
