@@ -1,21 +1,10 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-[ "$CC" = cc ] || { echo skipped; exit; }
+[ "$CC" = cc ] || skip
 
 # ASAN doesn't work with LD_PRELOAD
-nm mold-wrapper.so | grep -q '__[at]san_init' && { echo skipped; exit; }
+nm mold-wrapper.so | grep -q '__[at]san_init' && skip
 
 cat <<'EOF' | $CC -xc -c -o $t/a.o -
 #include <stdio.h>
@@ -61,5 +50,3 @@ chmod 755 $t/sh
 ./mold -run $t/sh $t/ld.lld --version | grep -q mold
 ./mold -run $t/sh $t/ld.gold --version | grep -q mold
 ./mold -run $t/sh $t/foo.ld --version | grep -q mold && false
-
-echo OK
