@@ -1,27 +1,14 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-echo 'int main() {}' | $CC -o /dev/null -xc - -static >& /dev/null || \
-  { echo skipped; exit; }
+test_cflags -static || skip
 
 if [ $MACHINE = x86_64 ]; then
   dialect=gnu2
 elif [ $MACHINE = aarch64 ]; then
   dialect=desc
 else
-  echo skipped
-  exit
+  skip
 fi
 
 cat <<EOF | $GCC -fPIC -mtls-dialect=$dialect -c -o $t/a.o -xc -
@@ -44,5 +31,3 @@ $QEMU $t/exe | grep -q 42
 
 $CC -B. -o $t/exe $t/a.o $t/b.o -static -Wl,-no-relax
 $QEMU $t/exe | grep -q 42
-
-echo OK
